'use strict';

const path = require('path');
const { exec } = require('child_process');
const async = require('async');
const jar = require('vnu-jar');
const chunkify = require('./chunkify.js');
const javadetect = require('./javadetect.js');

// chunkify's max characters
const MAX_CHARS = 5000;

// Increase child process buffer to accommodate large amounts of
// validation output. (the default is a paltry 200k)
// https://nodejs.org/api/child_process.html#child_process_child_process_exec_command_options_callback
const MAX_BUFFER = 20000 * 1024;

// Replace left/right quotation marks with normal quotation marks
function normalizeQuotationMarks(string) {
  if (string) {
    string = string.replace(/[\u201C\u201D]/g, '"');
  }

  return string;
}

// Parse and, if needed, normalize error messages from HttpClient to java -jar format
// java -jar: one object containing messages for all files
//   { messages: [{ message, type, url, ... }, ...] }
// HttpClient: one object per file, separated by a newline,
// each object containing messages for only that file
//   { messages: [{ message, type, ...}, ...], url }\n{ ... }
function parseErrorMessages(errors, config) {
  const parsed = JSON.parse(config.server ? `[${errors.trim().replace(/\n/g, ',')}]` : errors);
  let { messages } = parsed;

  if (config.server) {
    // Extract "messages" property from each object and set the url of each message.
    // This results in an array of arrays instead of array of objects, which is then
    // flattened by concatenation
    // eslint-disable-next-line unicorn/prefer-spread
    messages = [].concat(...messages, parsed.map(file => {
      return file.messages.map(message => {
        message.url = file.url;
        return message;
      });
    }));
  }

  return messages;
}

// Determine proper jarfile command and arguments
function javaCmd(java, chunk, config) {
  let args = '';

  if (config.server) {
    if (config.server.host) {
      args += ` -Dnu.validator.client.host=${config.server.host}`;
    }

    if (config.server.port) {
      args += ` -Dnu.validator.client.port=${config.server.port}`;
    }

    args += ' -Dnu.validator.client.out=json nu.validator.client.HttpClient';
  } else {
    args += ' --format json';
  }

  if (config.noLangDetect) {
    args += ' --no-langdetect';
  }

  const invoke = `${config.server ? '-cp' : '-jar'} "${jar}"${args}`;

  // Command to call java, increasing the default stack size for ia32 versions of the JRE
  // and using the default setting for x64 versions
  return `java ${java.arch === 'ia32' ? '-Xss512k ' : ''}${invoke} ${chunk}`;
}

function htmllint(config, done) {
  if (config.files.length === 0) {
    return done(null, []);
  }

  javadetect((err, java) => {
    if (err) {
      throw err;
    }

    const javaVersion = java.version.split('.').map(part => Number(part));

    if ((javaVersion[0] !== 1 && javaVersion[0] < 8) || (javaVersion[0] === 1 && javaVersion[1] < 8)) {
      throw new Error(`\nUnsupported Java version used: ${java.version}. Java 8 environment or up is required!`);
    }

    const files = config.files.map(file => path.normalize(file));

    async.mapSeries(chunkify(files, MAX_CHARS), (chunk, cb) => {
      exec(javaCmd(java, chunk, config), { maxBuffer: MAX_BUFFER }, (error, stdout, stderr) => {
        if (error && (error.code !== 1 || error.killed || error.signal)) {
          cb(error);
          return;
        }

        stderr = config.server ? stdout : stderr;
        let result = [];

        if (stderr) {
          try {
            result = parseErrorMessages(stderr, config);
          } catch (error_) {
            throw new Error(`${error_}\nInvalid input follows below:\n\n${stderr}`);
          }

          for (const message of result) {
            if (message.url) {
              message.file = path.relative('.', message.url.replace(path.sep === '\\' ? 'file:/' : 'file:', ''));
            }

            if (config.absoluteFilePathsForReporter) {
              message.file = path.resolve(message.file);
            }
          }

          if (config.ignore) {
            const ignore = Array.isArray(config.ignore) ? config.ignore : [config.ignore];

            result = result.filter(message => {
              // Iterate over the ignore rules and test the message against each rule.
              // A match should return false, which causes every() to return false and
              // the message to be filtered out.
              return ignore.every(currentValue => {
                if (currentValue instanceof RegExp) {
                  return !currentValue.test(message.message);
                }

                return normalizeQuotationMarks(currentValue) !== normalizeQuotationMarks(message.message);
              });
            });
          }
        }

        cb(null, result);
      });
    }, (error, results) => {
      if (error) {
        done(error);
        return;
      }

      let result = [];

      for (result of results) {
        result = [...result];
      }

      done(null, result.filter(item => config.errorlevels.includes(item.type)));
    });
  });
}

module.exports = htmllint;
